  &&lbl_ACC0, &&lbl_ACC1, &&lbl_ACC2, &&lbl_ACC3, &&lbl_ACC4, &&lbl_ACC5, &&lbl_ACC6, &&lbl_ACC7,
  &&lbl_ACC, &&lbl_PUSH,
  &&lbl_PUSHACC0, &&lbl_PUSHACC1, &&lbl_PUSHACC2, &&lbl_PUSHACC3,
  &&lbl_PUSHACC4, &&lbl_PUSHACC5, &&lbl_PUSHACC6, &&lbl_PUSHACC7,
  &&lbl_PUSHACC, &&lbl_POP, &&lbl_ASSIGN,
  &&lbl_ENVACC1, &&lbl_ENVACC2, &&lbl_ENVACC3, &&lbl_ENVACC4, &&lbl_ENVACC,
  &&lbl_PUSHENVACC1, &&lbl_PUSHENVACC2, &&lbl_PUSHENVACC3, &&lbl_PUSHENVACC4, &&lbl_PUSHENVACC,
  &&lbl_PUSH_RETADDR, &&lbl_APPLY, &&lbl_APPLY1, &&lbl_APPLY2, &&lbl_APPLY3,
  &&lbl_APPTERM, &&lbl_APPTERM1, &&lbl_APPTERM2, &&lbl_APPTERM3,
  &&lbl_RETURN, &&lbl_RESTART, &&lbl_GRAB,
  &&lbl_CLOSURE, &&lbl_CLOSUREREC,
  &&lbl_OFFSETCLOSUREM3, &&lbl_OFFSETCLOSURE0, &&lbl_OFFSETCLOSURE3, &&lbl_OFFSETCLOSURE,
  &&lbl_PUSHOFFSETCLOSUREM3, &&lbl_PUSHOFFSETCLOSURE0,
  &&lbl_PUSHOFFSETCLOSURE3, &&lbl_PUSHOFFSETCLOSURE,
  &&lbl_GETGLOBAL, &&lbl_PUSHGETGLOBAL, &&lbl_GETGLOBALFIELD, &&lbl_PUSHGETGLOBALFIELD, &&lbl_SETGLOBAL,
  &&lbl_ATOM0, &&lbl_ATOM, &&lbl_PUSHATOM0, &&lbl_PUSHATOM,
  &&lbl_MAKEBLOCK, &&lbl_MAKEBLOCK1, &&lbl_MAKEBLOCK2, &&lbl_MAKEBLOCK3, &&lbl_MAKEFLOATBLOCK,
  &&lbl_GETFIELD0, &&lbl_GETFIELD1, &&lbl_GETFIELD2, &&lbl_GETFIELD3, &&lbl_GETFIELD, &&lbl_GETFLOATFIELD,
  &&lbl_SETFIELD0, &&lbl_SETFIELD1, &&lbl_SETFIELD2, &&lbl_SETFIELD3, &&lbl_SETFIELD, &&lbl_SETFLOATFIELD,
  &&lbl_VECTLENGTH, &&lbl_GETVECTITEM, &&lbl_SETVECTITEM,
  &&lbl_GETBYTESCHAR, &&lbl_SETBYTESCHAR,
  &&lbl_BRANCH, &&lbl_BRANCHIF, &&lbl_BRANCHIFNOT, &&lbl_SWITCH, &&lbl_BOOLNOT,
  &&lbl_PUSHTRAP, &&lbl_POPTRAP, &&lbl_RAISE,
  &&lbl_CHECK_SIGNALS,
  &&lbl_C_CALL1, &&lbl_C_CALL2, &&lbl_C_CALL3, &&lbl_C_CALL4, &&lbl_C_CALL5, &&lbl_C_CALLN,
  &&lbl_CONST0, &&lbl_CONST1, &&lbl_CONST2, &&lbl_CONST3, &&lbl_CONSTINT,
  &&lbl_PUSHCONST0, &&lbl_PUSHCONST1, &&lbl_PUSHCONST2, &&lbl_PUSHCONST3, &&lbl_PUSHCONSTINT,
  &&lbl_NEGINT, &&lbl_ADDINT, &&lbl_SUBINT, &&lbl_MULINT, &&lbl_DIVINT, &&lbl_MODINT,
  &&lbl_ANDINT, &&lbl_ORINT, &&lbl_XORINT, &&lbl_LSLINT, &&lbl_LSRINT, &&lbl_ASRINT,
  &&lbl_EQ, &&lbl_NEQ, &&lbl_LTINT, &&lbl_LEINT, &&lbl_GTINT, &&lbl_GEINT,
  &&lbl_OFFSETINT, &&lbl_OFFSETREF, &&lbl_ISINT,
  &&lbl_GETMETHOD,
  &&lbl_BEQ, &&lbl_BNEQ,  &&lbl_BLTINT, &&lbl_BLEINT, &&lbl_BGTINT, &&lbl_BGEINT,
  &&lbl_ULTINT, &&lbl_UGEINT,
  &&lbl_BULTINT, &&lbl_BUGEINT,
  &&lbl_GETPUBMET, &&lbl_GETDYNMET,
  &&lbl_STOP,
  &&lbl_EVENT, &&lbl_BREAK,
  &&lbl_RERAISE, &&lbl_RAISE_NOTRACE,
  &&lbl_GETSTRINGCHAR,
  &&lbl_PERFORM, &&lbl_RESUME, &&lbl_RESUMETERM, &&lbl_REPERFORMTERM,
