/**************************************************************************/
/*                                                                        */
/*                                 OCaml                                  */
/*                                                                        */
/*             Xavier Leroy, projet Cristal, INRIA Rocquencourt           */
/*                                                                        */
/*   Copyright 2001 Institut National de Recherche en Informatique et     */
/*     en Automatique.                                                    */
/*                                                                        */
/*   All rights reserved.  This file is distributed under the terms of    */
/*   the GNU Lesser General Public License version 2.1, with the          */
/*   special exception on linking described in the file LICENSE.          */
/*                                                                        */
/**************************************************************************/

#ifndef CAML_PRINTEXC_H
#define CAML_PRINTEXC_H

#include "misc.h"
#include "mlvalues.h"

#ifdef __cplusplus
extern "C" {
#endif

CAMLextern char * caml_format_exception (value);

#ifdef __cplusplus
}
#endif

#ifdef CAML_INTERNALS
CAMLnoret void caml_fatal_uncaught_exception (value);
#endif /* CAML_INTERNALS */

#endif /* CAML_PRINTEXC_H */
